@@ -1,7 +1,6 @@
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 1996   Erich Boleyn  <erich@uruk.org>
- *  Copyright (C) 2000   Free Software Foundation, Inc.
+ *  Copyright (C) 2000, 2001   Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -68,6 +67,7 @@ int
 fat_mount (void)
 {
   struct fat_bpb bpb;
+  __u32 magic, first_fat;
   
   /* Check partition type for harddisk */
   if (((current_drive & 0x80) || (current_slice != 0))
@@ -162,6 +162,31 @@ fat_mount (void)
  	  > FAT_SUPER->fat_length))
     return 0;
   
+  /* kbs: Media check on first FAT entry [ported from PUPA] */
+
+  if (!devread(FAT_SUPER->fat_offset, 0,
+               sizeof(first_fat), (char *)&first_fat))
+    return 0;
+
+  if (FAT_SUPER->fat_size == 8)
+    {
+      first_fat &= 0x0fffffff;
+      magic = 0x0fffff00;
+    }
+  else if (FAT_SUPER->fat_size == 4)
+    {
+      first_fat &= 0x0000ffff;
+      magic = 0xff00;
+    }
+  else
+    {
+      first_fat &= 0x00000fff;
+      magic = 0x0f00;
+    }
+
+  if (first_fat != (magic | bpb.media))
+    return 0;
+
   FAT_SUPER->cached_fat = - 2 * FAT_CACHE_SIZE;
   return 1;
 }
@@ -176,7 +201,7 @@ fat_read (char *buf, int len)
   
   if (FAT_SUPER->file_cluster < 0)
     {
-      /* root directory for non-fat16 */
+      /* root directory for fat16 */
       size = FAT_SUPER->root_max - filepos;
       if (size > len)
  	size = len;
@@ -244,15 +269,11 @@ fat_read (char *buf, int len)
       if (size > len)
 	size = len;
       
-#ifndef STAGE1_5
       disk_read_func = disk_read_hook;
-#endif /* STAGE1_5 */
       
       devread(sector, offset, size, buf);
       
-#ifndef STAGE1_5
       disk_read_func = NULL;
-#endif /* STAGE1_5 */
       
       len -= size;
       buf += size;
